#pragma once
#include "../Auxiliary/bits.hpp"
#include "../jabyengine_defines.h"

namespace JabyEngine {
    namespace GPU {
        namespace internal {
            template<typename T, typename S>
            struct XYMovement {
                constexpr T& add(S dx, S dy) {
                    static_cast<T*>(this)->x += dx;
                    static_cast<T*>(this)->y += dy;
                    
                    return *static_cast<T*>(this);
                }

                constexpr T add(S dx, S dy) const {
                    return T::create(static_cast<const T*>(this)->x, static_cast<const T*>(this)->y).add(dx, dy);
                }

                constexpr T& add(const T& offset) {
                    return add(offset.x, offset.y);
                }

                constexpr T add(const T& offset) const {
                    return add(offset.x, offset.y);
                }

                constexpr T& sub(S dx, S dy) {
                    static_cast<T*>(this)->x -= dx;
                    static_cast<T*>(this)->y -= dy;
                    
                    return *static_cast<T*>(this);
                }

                constexpr T sub(S dx, S dy) const {
                    return T::create(static_cast<const T*>(this)->x, static_cast<const T*>(this)->y).sub(dx, dy);
                }

                constexpr T& sub(const T& offset) {
                    return sub(offset.x, offset.y);
                }

                constexpr T sub(const T& offset) const {
                    return sub(offset.x, offset.y);
                }

                constexpr T& move(S dx, S dy) {    
                    return this->add(dx, dy);
                }

                constexpr T move(S dx, S dy) const {
                    return T::create(static_cast<const T*>(this)->x, static_cast<const T*>(this)->y).move(dx, dy);
                }
            };
        }

        enum struct SemiTransparency {
            B_Half_add_F_Half = 0,
            B_add_F           = 1,
            B_sub_F           = 2,
            B_add_F_Quarter   = 3,
        };

        enum struct TexturePageColor {
            $4bit  = 0,
            $8bit  = 1,
            $15bit = 2,
        };

        struct Color24 {
            uint8_t red;
            uint8_t green;
            uint8_t blue;        

            static constexpr Color24 from_rgb(uint8_t r, uint8_t g, uint8_t b) {
                return Color24{r, g, b};
            }

            constexpr uint32_t raw() const {
                return ((this->blue << 16) | (this->green << 8) | this->red);
            }

            static constexpr Color24 Black() {
                return Color24(0, 0, 0);
            }

            static constexpr Color24 Grey() {
                return Color24(0x80, 0x80, 0x80);
            }

            static constexpr Color24 White() {
                return Color24(0xFF, 0xFF, 0xFF);
            }

            static constexpr Color24 Red(uint8_t base = 0xFF) {
                return Color24(base, 0x0, 0x0);
            }

            static constexpr Color24 Green(uint8_t base = 0xFF) {
                return Color24(0x0, base, 0x0);
            }

            static constexpr Color24 Blue(uint8_t base = 0xFF) {
                return Color24(0x0, 0x0, base);
            }

            static constexpr Color24 Yellow(uint8_t base = 0xFF) {
                return Color24(base, base, 0x0);
            }

            constexpr Color24 invert() const {
                return Color24(0xFF - this->red, 0xFF - this->green, 0xFF - this->blue);
            }
        };

        class Color {
        private:
            static constexpr auto RedRange            = BitRange::from_to(0, 4);
            static constexpr auto GreenRange          = BitRange::from_to(5, 9);
            static constexpr auto BlueRange           = BitRange::from_to(10, 14);
            static constexpr auto SemiTransperancyBit = Bit(15);

            uint16_t value;

        public:
            static constexpr Color from_rgb(uint8_t r, uint8_t g, uint8_t b) {
                return Color().set_red(r).set_green(g).set_blue(b);
            }

            static constexpr Color from(const Color24& color) {
                return Color::from_rgb(color.red, color.green, color.blue);
            }

            constexpr Color& set_red(uint8_t red) {
                this->value = bit::value::set_normalized(this->value, RedRange.with(red));
                return *this;
            }

            constexpr Color& set_green(uint8_t green) {
                this->value = bit::value::set_normalized(this->value, GreenRange.with(green));
                return *this;
            }

            constexpr Color& set_blue(uint8_t blue) {
                this->value = bit::value::set_normalized(this->value, BlueRange.with(blue));
                return *this;
            }
        };

        template<typename T>
        struct Position : public internal::XYMovement<Position<T>, T> {
            T x;
            T y;

            static constexpr Position create(T x, T y) {
                return Position{.x = x, .y = y};
            }
        };
        using PositionI8  = Position<int8_t>;
        using PositionU8  = Position<uint8_t>;
        using PositionI16 = Position<int16_t>;
        using PositionU16 = Position<uint16_t>;

        template<typename T>
        struct Size {
            T width;
            T height;

            static constexpr Size create(T w, T h) {
                return Size{w, h};
            }
        };
        typedef Size<int16_t>  SizeI16;
        typedef Size<uint16_t> SizeU16;
        // Type used for primitives
        typedef PositionI16 Vertex;

        template<typename T, typename S>
        static constexpr T tile_id_for(S id, S row_count, Size<S> size) {
            const auto x = id%row_count;
            const auto y = id/row_count;

            return T::create(static_cast<S>(size.width*x), static_cast<S>(size.height*y));
        }

        template<typename T, typename S>
        static constexpr T tile_id_for16(S id) {
            const auto x = id&0xF;
            const auto y = id >> 4;

            return T::create(static_cast<S>(x << 4), static_cast<S>(y << 4));
        }

        template<typename T>
        struct Area {
            Position<T> position;
            Size<T>     size;

            static constexpr Area create(Position<T> position, Size<T> size) {
                return Area{position, size};
            }

            static constexpr Area create(T position_x, T position_y, T size_width, T size_height) {
                return Area{Position<T>::create(position_x, position_y), Size<T>::create(size_width, size_height)};
            }

            constexpr Area centered() const {
                return Area(this->position.sub(this->size.width/2, this->size.height/2), this->size);
            }

            constexpr Position<T> get_bottom_left() const {
                return this->position.move(this->size.width, this->size.height);
            }
        };
        typedef Area<int16_t>  AreaI16;
        typedef Area<uint16_t> AreaU16;

        // Type used for primitives
        struct PageOffset : public internal::XYMovement<PageOffset, uint8_t> {
            union {
                uint8_t x;
                uint8_t u;
            };

            union {
                uint8_t y;
                uint8_t v;
            };

            static constexpr PageOffset create(uint8_t u, uint8_t v) {
                return PageOffset{.x = u, .y = v}; //< Activate x and y to use XYMovement during constexpr
            }

            static constexpr PageOffset from_tile_id(int16_t id, int16_t row_count, SizeI16 size) {
                return tile_id_for<PageOffset>(id, row_count, size);
            }
        };

        struct VertexColor {
            Vertex  position;
            Color24 color;

            static constexpr VertexColor create(Vertex vertex, Color24 color) {
                return VertexColor{vertex, color};
            }
        };

        struct ColorVertex {
                       Color24 color;
            alignas(2) Vertex  position;

            static constexpr ColorVertex create(Color24 color, Vertex vertex) {
                return ColorVertex{color, vertex};
            }
        };
    }
}